// SPDX-License-Identifier: MIT

use gumdrop::Options;
use ipnet::IpNet;
use once_cell::sync::OnceCell;
use tracing::Level;
use tracing_subscriber::prelude::*;

#[derive(Debug, Options)]
pub struct Args {
    #[options(
        no_short,
        help = "subnets to allow connection from. If unset, defaults to local segments"
    )]
    pub allowed_subnets: Vec<IpNet>,

    // up to 4 times only increment abos_web traces
    // more increment all traces for all crates
    #[options(count, help = "Increase verbosity (up to 7 times)")]
    verbose: u8,

    #[options(help = "print help message")]
    help: bool,
}

static ARGS: OnceCell<Args> = OnceCell::new();

fn init_logs() {
    let args = args();
    let abos_web_level = match args.verbose {
        0 => Level::WARN,
        1 => Level::INFO,
        2 => Level::DEBUG,
        _ => Level::TRACE,
    };
    let default_level = match args.verbose {
        i if i <= 3 => Level::ERROR,
        4 => Level::WARN,
        5 => Level::INFO,
        6 => Level::DEBUG,
        _ => Level::TRACE,
    };
    let filter = tracing_subscriber::filter::Targets::new()
        .with_target("abos_web", abos_web_level)
        .with_default(default_level);
    tracing_subscriber::registry()
        .with(tracing_subscriber::fmt::layer())
        .with(filter)
        .init();
}

pub fn args_init() {
    ARGS.set(Args::parse_args_default_or_exit())
        .expect("args init should only be called once");

    init_logs()
}

pub fn args() -> &'static Args {
    ARGS.get().expect("args init should have been called")
}
