// SPDX-License-Identifier: MIT

use axum::{http::StatusCode, Router};
use axum_sessions::{async_session::MemoryStore, SessionLayer};
use rand::Rng;
use std::net::SocketAddr;
use std::time::Duration;
use tower_http::services::ServeDir;

#[allow(unused_imports)]
use anyhow::{Context, Error, Result};

#[cfg(feature = "tls")]
use axum_server::tls_openssl::OpenSSLConfig;
#[cfg(feature = "rustls")]
use axum_server::tls_rustls::RustlsConfig;
#[cfg(any(feature = "tls", feature = "rustls"))]
use std::path::PathBuf;

mod args;
mod common;
mod connection;
mod container;
mod dhcp;
mod error;
mod ip_filter;
mod login;
mod nat;
mod nwstatbar;
mod status;
mod swu;
mod top;
mod vpn;
mod wlan;
mod wwan;
use connection::ConnectionRoutes;
use container::ContainerRoutes;
use dhcp::DhcpRoutes;
use login::LoginRoutes;
use nat::NatRoutes;
use nwstatbar::NWStatBarRoutes;
use status::StatusRoutes;
use swu::SwuRoutes;
use top::TopRoutes;
use vpn::VpnRoutes;
use wlan::WlanRoutes;
use wwan::WwanRoutes;

#[cfg(debug_assertions)]
const STATIC_RESOURCE_DIR: &str = "static";
#[cfg(all(any(feature = "tls", feature = "rustls"), debug_assertions))]
const TLS_DIR: &str = "tls";

#[cfg(not(debug_assertions))]
const STATIC_RESOURCE_DIR: &str = "/usr/share/abos-web/static";
#[cfg(all(any(feature = "tls", feature = "rustls"), not(debug_assertions)))]
const TLS_DIR: &str = "/etc/atmark/abos_web/tls";

async fn fallback_404() -> (StatusCode, &'static str) {
    (StatusCode::NOT_FOUND, "Page not found")
}

#[cfg(not(any(feature = "tls", feature = "rustls")))]
async fn tls_config() -> Result<()> {
    Err(Error::msg("not compiled with tls"))
}

#[cfg(feature = "tls")]
async fn tls_config() -> Result<OpenSSLConfig> {
    OpenSSLConfig::from_pem_file(
        PathBuf::from(TLS_DIR).join("cert.pem"),
        PathBuf::from(TLS_DIR).join("key.pem"),
    )
    .context("Could not initialize ssl context")
}

#[cfg(feature = "rustls")]
async fn tls_config() -> Result<RustlsConfig> {
    RustlsConfig::from_pem_file(
        PathBuf::from(TLS_DIR).join("cert.pem"),
        PathBuf::from(TLS_DIR).join("key.pem"),
    )
    .await
    .context("Could not initialize ssl context")
}

#[tokio::main]
async fn main() -> Result<()> {
    let store = MemoryStore::new();
    let secret = rand::thread_rng().gen::<[u8; 128]>();
    let session_layer = SessionLayer::new(store, &secret)
        .with_secure(false)
        .with_session_ttl(Some(Duration::from_secs(60 * 60 * 12)));
    common::Config::load().await?;
    login::password_throttle_init();
    args::args_init();

    let app = Router::new()
        .connection_routes()
        .container_routes()
        .dhcp_routes()
        .login_routes()
        .nat_routes()
        .nwstatbar_routes()
        .status_routes()
        .swu_routes()
        .top_routes()
        .vpn_routes()
        .wlan_routes()
        .wwan_routes()
        .nest_service("/static", ServeDir::new(STATIC_RESOURCE_DIR))
        .layer(session_layer)
        .route_layer(axum::middleware::from_fn(ip_filter::ipaddr_filter))
        .fallback(fallback_404);

    let addr: SocketAddr = "[::]:58080"
        .parse()
        .expect("Unable to parse socket address");
    tracing::debug!("listening on {}", addr);

    // fall back to plain http on error
    match tls_config().await {
        Err(e) => {
            tracing::info!(
                "Could not parse pem files, falling back to plain text: {}",
                e
            );
            axum_server::bind(addr)
                .serve(app.into_make_service_with_connect_info::<SocketAddr>())
                .await
                .unwrap();
        }
        #[allow(unused_variables)]
        Ok(config) => {
            #[cfg(feature = "tls")]
            axum_server::bind_openssl(addr, config)
                .serve(app.into_make_service_with_connect_info::<SocketAddr>())
                .await
                .unwrap();
            #[cfg(feature = "rustls")]
            axum_server::bind_rustls(addr, config)
                .serve(app.into_make_service_with_connect_info::<SocketAddr>())
                .await
                .unwrap();
        }
    }
    Ok(())
}
