// SPDX-License-Identifier: MIT

use anyhow::{Error, Result};
use askama::Template;
use axum::{
    response::{IntoResponse, Redirect},
    routing::get,
    Router,
};
use axum_sessions::extractors::ReadableSession;
use std::collections::HashMap;

use crate::common;
use crate::common::{
    container::ContainerInfo, html_template::HtmlTemplate, IptablesInfo, NatInfo,
    PortForwardingInfo,
};
use crate::error::PageResult;

#[derive(Template)]
#[template(path = "../src/status/templates/status.html")]
struct StatusTemplate {
    wwan_info: HashMap<&'static str, String>,
    wwan_addr: HashMap<&'static str, String>,
    wlan_addr: HashMap<&'static str, String>,
    ap_info: HashMap<&'static str, String>,
    lan_info: Vec<HashMap<&'static str, String>>,
    dhcp_info: Vec<HashMap<&'static str, String>>,
    nat_info: Vec<NatInfo>,
    pf_info: Vec<PortForwardingInfo>,
    container_info: ContainerInfo,
    vpn_info: HashMap<&'static str, String>,
    vpn_addr: HashMap<&'static str, String>,
}

pub trait StatusRoutes {
    fn status_routes(self) -> Self;
}

impl StatusRoutes for Router {
    fn status_routes(self) -> Self {
        self.route("/status", get(status))
    }
}

async fn status(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let (wwan_addr, wwan_info) = wwan_info().await?;
    let (wlan_addr, ap_info) = wlan_info().await?;
    let lan_info = lan_info().await?;
    let dhcp_info = dhcp_info().await?;
    let iptables_info = IptablesInfo::get().await?;
    let container_info = ContainerInfo::get().await?;
    let (vpn_addr, vpn_info) = vpn_info().await?;
    let template = StatusTemplate {
        wwan_addr,
        wwan_info,
        wlan_addr,
        ap_info,
        lan_info,
        dhcp_info,
        nat_info: iptables_info.nat,
        pf_info: iptables_info.port_forwarding,
        container_info,
        vpn_addr,
        vpn_info,
    };
    Ok(HtmlTemplate::new("状態一覧", template).into_response())
}

pub async fn wwan_info(
) -> Result<(HashMap<&'static str, String>, HashMap<&'static str, String>), Error> {
    let wwan_info = match common::wwan_info(common::WWAN_CON_NAME).await {
        Ok(wwan) => wwan,
        Err(e) => return Err(Error::msg(e)),
    };

    let interface = common::wwan_interface().unwrap_or("--");
    let wwan_addr = match common::current_address("ttyCommModem", interface).await {
        Ok(addr) => addr,
        Err(e) => return Err(Error::msg(e)),
    };
    Ok((wwan_addr, wwan_info))
}

pub async fn wlan_info(
) -> Result<(HashMap<&'static str, String>, HashMap<&'static str, String>), Error> {
    let interface = common::wlan_interface().unwrap_or("--");
    let args = &["get_wlan_con_ssid.sh", interface];
    let output = common::exec_command(args).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let v = stdout.trim().split(';').collect::<Vec<&str>>();
    let mut ssid = "";
    let mut security = "";
    if v.len() == 2 {
        ssid = if v[0].is_empty() { "--" } else { v[0] };
        security = if v[1].is_empty() { "--" } else { v[1] };
    }

    let mut wlan_addr = match common::current_address(interface, interface).await {
        Ok(addr) => addr,
        Err(e) => return Err(Error::msg(e)),
    };
    wlan_addr.insert("ssid", ssid.to_string());
    wlan_addr.insert("security", security.to_string());
    let ap_info = match common::ap_info().await {
        Ok(ap) => ap,
        Err(e) => return Err(Error::msg(e)),
    };

    Ok((wlan_addr, ap_info))
}

async fn lan_info() -> Result<Vec<HashMap<&'static str, String>>, Error> {
    let interface = "eth0";
    let lan_addr = match common::current_address(interface, interface).await {
        Ok(addr) => addr,
        Err(e) => return Err(Error::msg(e)),
    };
    let mut lan_addrs: Vec<HashMap<&'static str, String>> = Vec::new();
    lan_addrs.push(lan_addr);

    let exists_eth1 = common::exists_eth1();
    if exists_eth1 {
        let interface = "eth1";
        let lan_addr = match common::current_address(interface, interface).await {
            Ok(addr) => addr,
            Err(e) => return Err(Error::msg(e)),
        };
        lan_addrs.push(lan_addr);
    }
    Ok(lan_addrs)
}

async fn dhcp_info() -> Result<Vec<HashMap<&'static str, String>>, Error> {
    let interface = "eth0";
    let dhcp_info = match common::dhcp_info(interface).await {
        Ok(dhcp) => dhcp,
        Err(e) => return Err(Error::msg(e)),
    };
    let mut dhcp_infos: Vec<HashMap<&'static str, String>> = Vec::new();
    if !dhcp_info.is_empty() {
        dhcp_infos.push(dhcp_info);
    }
    let exists_eth1 = common::exists_eth1();
    if exists_eth1 {
        let interface = "eth1";
        let dhcp_info = match common::dhcp_info(interface).await {
            Ok(dhcp) => dhcp,
            Err(e) => return Err(Error::msg(e)),
        };
        if !dhcp_info.is_empty() {
            dhcp_infos.push(dhcp_info);
        }
    }
    let exists_br_ap = common::exists_br_ap();
    if exists_br_ap {
        let dhcp_info = match common::dhcp_info(common::AP_BRIDGE_INTERFACE).await {
            Ok(dhcp) => dhcp,
            Err(e) => return Err(Error::msg(e)),
        };
        if !dhcp_info.is_empty() {
            dhcp_infos.push(dhcp_info);
        }
    }
    Ok(dhcp_infos)
}

pub async fn vpn_info(
) -> Result<(HashMap<&'static str, String>, HashMap<&'static str, String>), Error> {
    let args = &["vpn_info.sh"];
    let output = common::exec_command(args).await?;
    let setting_name = String::from_utf8_lossy(&output.stdout).trim().to_string();
    let mut vpn_info = HashMap::new();
    vpn_info.insert("setting_name", setting_name);

    let interface = "tun0";
    let vpn_addr = match common::current_address(interface, interface).await {
        Ok(addr) => addr,
        Err(e) => return Err(Error::msg(e)),
    };
    Ok((vpn_addr, vpn_info))
}
