// SPDX-License-Identifier: MIT

use anyhow::{Context, Result};
use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::ReadableSession;
use ipnet::Ipv4Net;
use serde::Deserialize;
use std::{collections::HashMap, net::Ipv4Addr, str::FromStr};

use crate::common;
use crate::common::html_template::HtmlTemplate;
use crate::error::{ErrorStringResult, PageResult};

#[derive(Deserialize)]
struct WlanParam {
    essid: String,
    other_essid: String,
    password: String,
    security: String,
    method: String,
    ip_addr: String,
    netmask: String,
    gateway_addr: String,
    p_dns_addr: String,
    s_dns_addr: String,
}

#[derive(Deserialize, Debug)]
struct WlanSaved {
    name: String,
    ssid: String,
    autoconnect: String,
}

#[derive(Deserialize, Debug)]
struct WlanChosen {
    con_uuid: String, // TODO: currently connection.id (name)
}

#[derive(Deserialize, Debug)]
struct APParam {
    bridge_addr: String,
    netmask: String,
    hw_mode: String,
    channel: String,
    ssid: String,
    wpa_passphrase: String,
}

#[derive(Deserialize, Debug)]
struct APDelParam {}

#[derive(Template)]
#[template(path = "../src/wlan/templates/wlan.html")]
struct WlanTemplate<'a> {
    ssid_list: Vec<&'a str>,
    wlan_saved: Vec<WlanSaved>,
    wlan_con_ssid: String,
    security: String,
    current_ip_addr: String,
    current_gw_addr: String,
    current_netmask: String,
    interface: String,
    ap_info: HashMap<&'static str, String>,
    ap_br_interface: String,
    has_wlan: bool,
}

pub trait WlanRoutes {
    fn wlan_routes(self) -> Self;
}

impl WlanRoutes for Router {
    fn wlan_routes(self) -> Self {
        self.route("/wlan", get(wlan))
            .route("/wlan_scan", get(wlan_scan))
            .route("/wlan_setup", post(wlan_setup))
            .route("/wlan_save", post(wlan_save))
            .route("/wlan_delete", post(wlan_delete))
            .route("/wlan_connect", post(wlan_connect))
            .route("/wlan_disconnect", post(wlan_disconnect))
            .route("/ap_setup", post(ap_setup))
            .route("/ap_delete", post(ap_delete))
    }
}

async fn wlan(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }
    let output = common::exec_command(&["get_ssid_list.sh", "--no-rescan"]).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let ssid_list = stdout.lines().map(|s| s.trim()).collect::<Vec<&str>>();

    let interface = common::wlan_interface().context("Could not find wlan network interface.")?;
    let args = &["get_wlan_con_ssid.sh", interface];
    let output = common::exec_command(args).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let v = stdout.trim().split(';').collect::<Vec<&str>>();
    let mut ssid = "";
    let mut security = "";
    if v.len() == 2 {
        ssid = if v[0].is_empty() { "--" } else { v[0] };
        security = if v[1].is_empty() { "--" } else { v[1] };
    }

    let interface = common::wlan_interface().unwrap_or("--");
    let current_addr = common::current_address(interface, interface).await?;
    let ap_info = common::ap_info().await?;

    let args = &["get_wlan_saved_con.sh"];
    let output = common::exec_command(args).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let wlan_saved = stdout
        .lines()
        .map(|s| s.trim().split(';').collect::<Vec<&str>>())
        .map(|a| WlanSaved {
            name: a.first().unwrap_or(&"--").to_string(),
            ssid: a.get(1).unwrap_or(&"--").to_string(),
            autoconnect: a.get(2).unwrap_or(&"--").to_string(),
        })
        .collect();

    let template = WlanTemplate {
        ssid_list,
        wlan_saved,
        wlan_con_ssid: ssid.to_string(),
        security: security.to_string(),
        current_ip_addr: current_addr["current_ip_addr"].to_owned(),
        current_netmask: current_addr["current_netmask"].to_owned(),
        current_gw_addr: current_addr["current_gw_addr"].to_owned(),
        interface: interface.to_string(),
        ap_info,
        ap_br_interface: common::AP_BRIDGE_INTERFACE.to_string(),
        has_wlan: common::wlan_interface().is_some(),
    };
    Ok(HtmlTemplate::new("WLAN設定", template).into_response())
}

async fn wlan_scan(session: ReadableSession) -> ErrorStringResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }
    let output = common::exec_command(&["get_ssid_list.sh", "--rescan"]).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let ssid_list = stdout.lines().map(|s| s.trim()).collect::<Vec<&str>>();
    let list = serde_json::to_string(&ssid_list).context("Could not format json.")?;
    let mut res = list.into_response();
    res.headers_mut()
        .insert("content-type", "application/json".parse().unwrap());
    Ok(res)
}

fn wlan_save_args(args: &mut Vec<String>, wlan_params: WlanParam) -> Result<()> {
    let interface = common::wlan_interface().context("Could not find wlan network interface.")?;
    let mut essid = wlan_params.essid;
    if essid == "other" {
        essid = wlan_params.other_essid;
    }
    args.push(essid);
    args.push(wlan_params.security);
    args.push(wlan_params.password);
    args.push(interface.to_string());
    args.push(wlan_params.method.clone());

    if wlan_params.method == "manual" {
        let netmask = if wlan_params.netmask.is_empty() {
            "255.255.255.255"
        } else {
            &wlan_params.netmask
        };
        let ip_addr = Ipv4Addr::from_str(&wlan_params.ip_addr)
            .with_context(|| format!("Could not parse addr: {}", wlan_params.ip_addr))?;
        let netmask = Ipv4Addr::from_str(netmask)
            .with_context(|| format!("Could not parse network: {}", netmask))?;
        let net = Ipv4Net::with_netmask(ip_addr, netmask)
            .with_context(|| format!("Could not define network with {} / {}", ip_addr, netmask))?;

        args.push("-a".to_string());
        args.push(net.to_string());
        if !wlan_params.gateway_addr.is_empty() {
            args.push("-g".to_string());
            args.push(wlan_params.gateway_addr);
        }
        if !wlan_params.p_dns_addr.is_empty() {
            let mut dns = wlan_params.p_dns_addr;
            if !wlan_params.s_dns_addr.is_empty() {
                dns = format!("{},{}", dns, wlan_params.s_dns_addr);
            }
            args.push("-d".to_string());
            args.push(dns);
        }
    }
    Ok(())
}

async fn wlan_setup(session: ReadableSession, Form(wlan_param): Form<WlanParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let mut args = vec!["wlan_setup.sh".to_string()];
    wlan_save_args(&mut args, wlan_param)?;
    args.push("--connect".to_string());

    common::exec_command(&args).await?;
    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_save(session: ReadableSession, Form(wlan_param): Form<WlanParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let mut args = vec!["wlan_setup.sh".to_string()];
    wlan_save_args(&mut args, wlan_param)?;

    common::exec_command(&args).await?;
    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_delete(session: ReadableSession, Form(saved_con): Form<WlanChosen>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["wlan_delete.sh", &saved_con.con_uuid];

    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_connect(session: ReadableSession, Form(saved_con): Form<WlanChosen>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["wlan_connect.sh", &saved_con.con_uuid, "up"];

    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_disconnect(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let interface = common::wlan_interface().unwrap_or("--");
    let args = &["wlan_disconnect.sh", interface];

    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn ap_setup(session: ReadableSession, Form(ap_param): Form<APParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let netmask = if ap_param.netmask.is_empty() {
        "255.255.255.255"
    } else {
        &ap_param.netmask
    };
    let ip_addr = Ipv4Addr::from_str(&ap_param.bridge_addr)
        .with_context(|| format!("Could not parse addr: {}", &ap_param.bridge_addr))?;
    let netmask = Ipv4Addr::from_str(netmask)
        .with_context(|| format!("Could not parse network: {}", netmask))?;
    let net = Ipv4Net::with_netmask(ip_addr, netmask)
        .with_context(|| format!("Could not define network with {} / {}", ip_addr, netmask))?;
    let interface = common::ap_interface().context("Could not find wlan network interface.")?;

    let args = &[
        "ap_setup.sh",
        common::AP_BRIDGE_NAME,
        &net.to_string(),
        &ap_param.hw_mode,
        &ap_param.channel,
        &ap_param.ssid,
        &ap_param.wpa_passphrase,
        interface,
    ];
    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn ap_delete(session: ReadableSession, Form(_): Form<APDelParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let interface = common::ap_interface().context("Could not find wlan network interface.")?;

    let args = &["ap_delete.sh", "abos_web_br_ap", interface];
    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}
