#!/bin/sh
# SPDX-License-Identifier: MIT

# This scripts install a local cargo toolchain in the test directory,
# builds and install abos-web on the target machine and then
# runs tests in the official playwright container
#
# For developement, it's recommended to install dependencies directly
# and run either utils/update_armadillo.sh for build/update and
# tests/run_playwright.sh to run a locally installed version of
# this test.

set -x

error() {
        printf "%s\n" "$@" >&2
        exit 1
}

cleanup() {
	# remove test artifacts from previous runs
	rm -rf playwright/screenshots playwright/test-results
}

setup() {
	export RUSTUP_HOME="$PWD/bins/rustup"
	export CARGO_HOME="$PWD/bins/cargo"
	export PATH="$PWD/bins/cargo/bin:$PATH"
	if command -v jq >/dev/null; then
		PLAYWRIGHT_VERSION=$(jq -r '.packages."node_modules/@playwright/test".version' < playwright/package-lock.json)
	else
		# ugly grep...
		PLAYWRIGHT_VERSION=$(grep -A 1 'node_modules/@playwright/test' playwright/package-lock.json \
					| sed -ne 's/.*version": "\([^"]*\)".*/\1/p')
	fi
	[ -n "$PLAYWRIGHT_VERSION" ] || error "Could not get playwright version from playwright/package-lock.json"
	PLAYWRIGHT_CTR="mcr.microsoft.com/playwright:v$PLAYWRIGHT_VERSION-jammy"

	if ! docker image inspect "$PLAYWRIGHT_CTR" >/dev/null 2>&1; then
		docker pull "$PLAYWRIGHT_CTR" || exit
	fi


	if ! [ -e bins/rustup ]; then
		# cross cannot run in a container because it needs to run its own container...
		# so we install our own rustup and everything in workspace
		mkdir -p bins/xargo
		curl https://sh.rustup.rs -sSf | sh -s -- --profile minimal --no-modify-path --target aarch64-unknown-linux-musl -y || exit
	fi
	if ! [ -e bins/cargo/bin/cross ]; then
		cargo install cross
	fi

	case "$armadillo" in
	*%*%*) error "MUltiple % signs not handled";;
	*%*)
		# double % for ssh config
		armadillo="${armadillo%\%*}%%${armadillo#*%}"
		;;
	esac
	cat > bins/ssh_config <<EOF
Host g4
	Hostname $armadillo
	User root
	BatchMode yes
	UserKnownHostsFile "$PWD/bins/ssh_known_hosts"
	StrictHostKeyChecking accept-new
EOF
	rm -f bins/ssh_known_hosts

	case "$armadillo" in
	fe80*)
		# browsers can't use local-link addresses yet... get ipv4.
		armadillo=$(ssh -F bins/ssh_config g4 ip -4 a show dev eth0 | sed -ne 's/.*inet \([0-9]*\.[0-9]*\.[0-9]*\.[0-9]*\).*/\1/p')
		[ -n "$armadillo" ] || error "Could not resolve address for armadillo"
		;;
	esac
}

update_armadillo() {
	echo 'abos-web-admin:abosweb_autotest' | ssh -F "$PWD/bins/ssh_config" g4 chpasswd || exit
	../utils/update_armadillo.sh g4 sshconf "$PWD/bins/ssh_config" || exit
}

run_tests() {
	cat > bins/.env <<EOF
ABOSWEB_BASE=https://$armadillo:58080
ABOSWEB_PASSWORD=abosweb_autotest
CI=1
EOF
	docker run --rm \
		-v "$PWD/playwright:/tests" -w /tests \
		--env-file $PWD/bins/.env \
		"$PLAYWRIGHT_CTR" \
		sh -c 'npm ci && npx playwright test'
}

main() {
	[ "$#" = 1 ] && [ "$1" != "-h" ] || error "Usage: $0 <armadillo_address>"
	armadillo="$1"

	cd "$(dirname "$0")" || exit
	cleanup
	setup
	update_armadillo
	run_tests
}

main "$@"
